from pathlib import Path
import imageio.v3 as iio
from pycvt import draw_bounding_boxes, generate_sliding_windows

from cvmd import detect_with_windows
from cvmd import Yolov11Segment
from .utils import make_mosaic


def main():
    img_dir = "/home/user/worksapce/cvmd/temp/coco128/images/train2017"
    mosaic_path = "/home/user/worksapce/cvmd/temp/mosaic_64.jpg"
    
    if not Path(mosaic_path).exists():
        print("Generating mosaic...")
        make_mosaic(
            img_dir=img_dir,
            num_images=64,
            thumb_size=(256, 256),
            seed=0,
            out_path=mosaic_path
        )

    image = iio.imread(mosaic_path, mode="RGB")
    windows = generate_sliding_windows(
        im_shape=image.shape[:2],
        ws=(800, 800),
        s=(200, 200),
    )
    
    weights = "/home/user/worksapce/cvmd/temp/yolo11l-seg.torchscript"
    if not Path(weights).exists():
        print(f"Weights not found: {weights}")
        return

    model = Yolov11Segment(
        weights=weights,
        device="cuda",
        conf=0.25,
        iou=0.45,
        imgsz=640,
        half=True,
    )
    model.load_model()
    
    detected = detect_with_windows(
        image=image,
        windows=windows,
        model=model,
        merge=True,
        merge_iou=0.001
    )
    
    print(f"Detected {detected.shape[0]} objects in total.")
    
    plot_im = draw_bounding_boxes(
        image=image,
        boxes=detected[..., :4],
        labels=[
            f"{c}-{s:.2f}"
            for c, s in zip(detected[..., 5].astype(int), detected[..., 4])
        ],
        line_width=3,
    )
    save_dir = Path("runs/windows_test")
    save_dir.mkdir(parents=True, exist_ok=True)
    iio.imwrite(save_dir / "detect_with_windows_result.jpg", plot_im)


if __name__ == "__main__":
    main()



from cvmd.registry import register_model
import numpy as np
from .v5ops import process_mask_native, non_max_suppression
from .yolov8 import Yolov8Detect, Yolov8Segment
from .ops import scale_boxes


@register_model("yolov5", "yolov5det", "yolov5detect")
class Yolov5Detect(Yolov8Detect):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __post_process__(self, pred, *args, **kwds):
        pred = non_max_suppression(
            pred, self.conf, self.iou, classes=self.classes
        )[0]
        pred[:, :4] = scale_boxes(self.imgsz, pred[:, :4], args[0].shape).round()
        return pred.cpu().numpy()


@register_model("yolov5seg", "yolov5segment")
class Yolov5Segment(Yolov8Segment):
    def __init__(self, *args, **kwargs):
        self.mask_thr = kwargs.pop("mask_thr", 0.6)
        self.nm = kwargs.get("nm", None)
        super().__init__(*args, **kwargs)

    def __post_process__(self, pred, *args, **kwds):
        pred, proto = pred

        proto = proto[0]

        if self.nm is None:
            self.nm = proto.shape[0]
        pred = non_max_suppression(
            pred,
            self.conf,
            self.iou,
            classes=self.classes,
            nm=self.nm,
        )[0]

        if len(pred) == 0:
            return np.zeros((0, 6), dtype=np.float32), np.zeros(
                (0, *args[0].shape[:2]), dtype=bool
            )

        pred[:, :4] = scale_boxes(self.imgsz, pred[:, :4], args[0].shape).round()
        masks = process_mask_native(proto, pred[:, 6:], pred[:, :4], args[0].shape[:2])
        masks = masks > self.mask_thr
        return pred[:, :6].cpu().numpy(), masks.cpu().numpy()



from cvmd.registry import register_model
from .yolov8 import Yolov8Detect, Yolov8Segment

@register_model("yolov11", "yolov11det", "yolov11detect")
class Yolov11Detect(Yolov8Detect):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

@register_model("yolov11seg", "yolov11segment")
class Yolov11Segment(Yolov8Segment):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from pathlib import Path
from typing import Any, Dict

from tqdm import tqdm
from cvmd import YoloInitKwargs
from cvmd import IMAGE_EXTS
import imageio.v3 as iio

from cvmd.utils.ray_infer import InferActor, ray_infer_iter
from pycvt import save_yolo_annotations, draw_bounding_boxes, xyxy2xywhn


# must can be pickled
def handler(
    task,
    model_config: Dict[str, Any],
    runs_config: Dict[str, Any],
    *args,
    **kwds,
) -> Dict[str, Any]:

    # print(f"Handling task: {task}")
    # print(f"Model config inside handler: {model_config}")
    # print(f"Runs config inside handler: {runs_config}")
    # print(f"Additional args: {args}, kwds: {kwds}")

    model = model_config["model"]
    out_dir = runs_config.get("save_dir", "runs/ray_runs_default")
    img_exts = IMAGE_EXTS
    folder_path = Path(task)
    save_dir = Path(runs_config.get("save_dir", out_dir)) / folder_path.name
    save_dir.mkdir(parents=True, exist_ok=True)

    # 获取当前进程ID
    import os

    pid = os.getpid()

    save_dir = save_dir / f"pid_{pid}"
    save_dir.mkdir(parents=True, exist_ok=True)

    save_plot = runs_config.get("save_plot", False)
    save_plot_dir = save_dir / "plots"
    if save_plot:
        save_plot_dir.mkdir(parents=True, exist_ok=True)

    save_txt = runs_config.get("save_txt", False)
    save_txt_dir = save_dir / "labels"
    if save_txt:
        save_txt_dir.mkdir(parents=True, exist_ok=True)

    save_source = runs_config.get("save_source", False)
    save_source_dir = save_dir / "source"
    if save_source:
        save_source_dir.mkdir(parents=True, exist_ok=True)

    results = []
    for img_path in folder_path.iterdir():
        if img_path.suffix.lower() not in img_exts:
            continue
        img = iio.imread(img_path.as_posix(), mode="RGB")
        pred = model(img)
        if isinstance(pred, tuple):
            pred = pred[0]  # for yolov11seg returning (pred, masks)
        cls = pred[:, 5]
        boxes = pred[:, :4]
        scores = pred[:, 4]
        
        if len(boxes) == 0:
            results.append((img_path.name, cls, boxes, scores))
            continue

        if save_txt:
            xywhn_boxes = xyxy2xywhn(boxes, img.shape[1], img.shape[0])
            save_yolo_annotations(
                file_path=save_txt_dir / f"{img_path.stem}.txt",
                cls=cls,
                bboxes=xywhn_boxes,
            )
        if save_plot:
            plot_img = draw_bounding_boxes(
                image=img,
                boxes=boxes,
                labels=[f"{int(c)}_{s:.2f}" for c, s in zip(cls, scores)],
                line_width=2
            )
            save_path = save_plot_dir / img_path.name
            save_path = save_path.with_suffix(".jpg")
            iio.imwrite(save_path.as_posix(), plot_img, quality=95)
        if save_source:
            source_path = save_source_dir / img_path.name
            if not source_path.exists():
                os.link(img_path.as_posix(), source_path.as_posix())

        results.append((img_path.name, cls, boxes, scores))
    return {"folder": folder_path.name, "results": results}


def main():

    data_root = "/home/user/worksapce/cvmd/temp/coco128/images/train2017"

    mode_config = YoloInitKwargs(
        model_name="yolov11seg",
        weights="/home/user/worksapce/cvmd/temp/yolo11l-seg.torchscript",
        device="cuda",
        conf=0.25,
        iou=0.9,
        imgsz=640,
        half=True,
        classes=list(range(40)),
    )

    runs_config = {
        "save_dir": "/home/user/worksapce/cvmd/runs/ray_runs3",
        "save_txt": True,
        "save_plot": True,
        "save_source": True,
    }

    tasks = [data_root] * 100  # 这里 task 可以是 folder path，也可以是 dict

    # ============================================
    # Start Ray inference
    # ============================================

    it = ray_infer_iter(
        InferActor,
        tasks,
        actor_kwargs={
            "runs_config": runs_config,
            "model_config": mode_config,
            "handler": handler,
        },
        remote_method="infer",
        # remote_args=(out_dir,), # if needed, may be use runs_config or model_config
    )

    # ============================================
    # Collect results （not ordered）
    # ============================================

    with tqdm(total=len(tasks), desc="infer folders", unit="folder") as pbar:
        for r in it:
            pbar.update(1)


if __name__ == "__main__":
    main()



import torch
from cvmd.registry import register_model
from cvmd.yolo.yolov8 import Yolov8Detect
import numpy as np
import torchvision.transforms as T
from cvmd.yolo.ops import xywhn2xyxy


@register_model("detr", "detrdetect", "DETR")
class DETR(Yolov8Detect):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.transform = T.Compose(
            [
                T.ToTensor(),
                T.Resize(self.imgsz),
                T.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
            ]
        )

    def __pre_process__(self, *args, **kwds):
        im0 = args[0]
        x = self.transform(im0).unsqueeze(0).to(self.device)  # 1CHW
        x = x.to(self._model_dtype)
        return x

    def __post_process__(self, preds, *args, **kwds):
        im0 = args[0]
        logits = preds["pred_logits"][0]
        boxes = preds["pred_boxes"][0]
        probs = logits.softmax(-1)
        scores, labels = probs[..., :-1].max(-1)

        keep = scores > self.conf
        
        if self.classes is not None:
            classes_t = torch.as_tensor(self.classes, device=labels.device)
            keep = keep & torch.isin(labels, classes_t)
                
        scores = scores[keep]
        labels = labels[keep]
        boxes = boxes[keep]

        if boxes.numel() == 0:
            return np.zeros((0, 6), dtype=np.float32)

        h, w = im0.shape[:2]
        xyxy = xywhn2xyxy(boxes, w, h)  # alternative way

        xyxy[:, 0::2].clamp_(min=0, max=w)
        xyxy[:, 1::2].clamp_(min=0, max=h)

        # Concatenate to (x1, y1, x2, y2, conf, cls)
        results = torch.cat(
            [xyxy, scores.unsqueeze(-1), labels.unsqueeze(-1).float()], dim=-1
        )

        return results.cpu().numpy().astype(np.float32)



from cvmd.registry import register_model


@register_model("deformabledetr", "deformable_detr")
class DeformableDETR:

    def __init__(self,*args,**kwargs):
        
        raise NotImplementedError("DeformableDETR is not yet implemented.")



# Ultralytics 🚀 AGPL-3.0 License - https://ultralytics.com/license
"""Model validation metrics."""

import math
import warnings
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import torch



OKS_SIGMA = (
    np.array([0.26, 0.25, 0.25, 0.35, 0.35, 0.79, 0.79, 0.72, 0.72, 0.62, 0.62, 1.07, 1.07, 0.87, 0.87, 0.89, 0.89])
    / 10.0
)


def bbox_ioa(box1: np.ndarray, box2: np.ndarray, iou: bool = False, eps: float = 1e-7) -> np.ndarray:
    """
    Calculate the intersection over box2 area given box1 and box2.

    Args:
        box1 (np.ndarray): A numpy array of shape (N, 4) representing N bounding boxes in x1y1x2y2 format.
        box2 (np.ndarray): A numpy array of shape (M, 4) representing M bounding boxes in x1y1x2y2 format.
        iou (bool, optional): Calculate the standard IoU if True else return inter_area/box2_area.
        eps (float, optional): A small value to avoid division by zero.

    Returns:
        (np.ndarray): A numpy array of shape (N, M) representing the intersection over box2 area.
    """
    # Get the coordinates of bounding boxes
    b1_x1, b1_y1, b1_x2, b1_y2 = box1.T
    b2_x1, b2_y1, b2_x2, b2_y2 = box2.T

    # Intersection area
    inter_area = (np.minimum(b1_x2[:, None], b2_x2) - np.maximum(b1_x1[:, None], b2_x1)).clip(0) * (
        np.minimum(b1_y2[:, None], b2_y2) - np.maximum(b1_y1[:, None], b2_y1)
    ).clip(0)

    # Box2 area
    area = (b2_x2 - b2_x1) * (b2_y2 - b2_y1)
    if iou:
        box1_area = (b1_x2 - b1_x1) * (b1_y2 - b1_y1)
        area = area + box1_area[:, None] - inter_area

    # Intersection over box2 area
    return inter_area / (area + eps)


def box_iou(box1: torch.Tensor, box2: torch.Tensor, eps: float = 1e-7) -> torch.Tensor:
    """
    Calculate intersection-over-union (IoU) of boxes.

    Args:
        box1 (torch.Tensor): A tensor of shape (N, 4) representing N bounding boxes in (x1, y1, x2, y2) format.
        box2 (torch.Tensor): A tensor of shape (M, 4) representing M bounding boxes in (x1, y1, x2, y2) format.
        eps (float, optional): A small value to avoid division by zero.

    Returns:
        (torch.Tensor): An NxM tensor containing the pairwise IoU values for every element in box1 and box2.

    References:
        https://github.com/pytorch/vision/blob/main/torchvision/ops/boxes.py
    """
    # NOTE: Need .float() to get accurate iou values
    # inter(N,M) = (rb(N,M,2) - lt(N,M,2)).clamp(0).prod(2)
    (a1, a2), (b1, b2) = box1.float().unsqueeze(1).chunk(2, 2), box2.float().unsqueeze(0).chunk(2, 2)
    inter = (torch.min(a2, b2) - torch.max(a1, b1)).clamp_(0).prod(2)

    # IoU = inter / (area1 + area2 - inter)
    return inter / ((a2 - a1).prod(2) + (b2 - b1).prod(2) - inter + eps)


def bbox_iou(
    box1: torch.Tensor,
    box2: torch.Tensor,
    xywh: bool = True,
    GIoU: bool = False,
    DIoU: bool = False,
    CIoU: bool = False,
    eps: float = 1e-7,
) -> torch.Tensor:
    """
    Calculate the Intersection over Union (IoU) between bounding boxes.

    This function supports various shapes for `box1` and `box2` as long as the last dimension is 4.
    For instance, you may pass tensors shaped like (4,), (N, 4), (B, N, 4), or (B, N, 1, 4).
    Internally, the code will split the last dimension into (x, y, w, h) if `xywh=True`,
    or (x1, y1, x2, y2) if `xywh=False`.

    Args:
        box1 (torch.Tensor): A tensor representing one or more bounding boxes, with the last dimension being 4.
        box2 (torch.Tensor): A tensor representing one or more bounding boxes, with the last dimension being 4.
        xywh (bool, optional): If True, input boxes are in (x, y, w, h) format. If False, input boxes are in
                               (x1, y1, x2, y2) format.
        GIoU (bool, optional): If True, calculate Generalized IoU.
        DIoU (bool, optional): If True, calculate Distance IoU.
        CIoU (bool, optional): If True, calculate Complete IoU.
        eps (float, optional): A small value to avoid division by zero.

    Returns:
        (torch.Tensor): IoU, GIoU, DIoU, or CIoU values depending on the specified flags.
    """
    # Get the coordinates of bounding boxes
    if xywh:  # transform from xywh to xyxy
        (x1, y1, w1, h1), (x2, y2, w2, h2) = box1.chunk(4, -1), box2.chunk(4, -1)
        w1_, h1_, w2_, h2_ = w1 / 2, h1 / 2, w2 / 2, h2 / 2
        b1_x1, b1_x2, b1_y1, b1_y2 = x1 - w1_, x1 + w1_, y1 - h1_, y1 + h1_
        b2_x1, b2_x2, b2_y1, b2_y2 = x2 - w2_, x2 + w2_, y2 - h2_, y2 + h2_
    else:  # x1, y1, x2, y2 = box1
        b1_x1, b1_y1, b1_x2, b1_y2 = box1.chunk(4, -1)
        b2_x1, b2_y1, b2_x2, b2_y2 = box2.chunk(4, -1)
        w1, h1 = b1_x2 - b1_x1, b1_y2 - b1_y1 + eps
        w2, h2 = b2_x2 - b2_x1, b2_y2 - b2_y1 + eps

    # Intersection area
    inter = (b1_x2.minimum(b2_x2) - b1_x1.maximum(b2_x1)).clamp_(0) * (
        b1_y2.minimum(b2_y2) - b1_y1.maximum(b2_y1)
    ).clamp_(0)

    # Union Area
    union = w1 * h1 + w2 * h2 - inter + eps

    # IoU
    iou = inter / union
    if CIoU or DIoU or GIoU:
        cw = b1_x2.maximum(b2_x2) - b1_x1.minimum(b2_x1)  # convex (smallest enclosing box) width
        ch = b1_y2.maximum(b2_y2) - b1_y1.minimum(b2_y1)  # convex height
        if CIoU or DIoU:  # Distance or Complete IoU https://arxiv.org/abs/1911.08287v1
            c2 = cw.pow(2) + ch.pow(2) + eps  # convex diagonal squared
            rho2 = (
                (b2_x1 + b2_x2 - b1_x1 - b1_x2).pow(2) + (b2_y1 + b2_y2 - b1_y1 - b1_y2).pow(2)
            ) / 4  # center dist**2
            if CIoU:  # https://github.com/Zzh-tju/DIoU-SSD-pytorch/blob/master/utils/box/box_utils.py#L47
                v = (4 / math.pi**2) * ((w2 / h2).atan() - (w1 / h1).atan()).pow(2)
                with torch.no_grad():
                    alpha = v / (v - iou + (1 + eps))
                return iou - (rho2 / c2 + v * alpha)  # CIoU
            return iou - rho2 / c2  # DIoU
        c_area = cw * ch + eps  # convex area
        return iou - (c_area - union) / c_area  # GIoU https://arxiv.org/pdf/1902.09630.pdf
    return iou  # IoU


def mask_iou(mask1: torch.Tensor, mask2: torch.Tensor, eps: float = 1e-7) -> torch.Tensor:
    """
    Calculate masks IoU.

    Args:
        mask1 (torch.Tensor): A tensor of shape (N, n) where N is the number of ground truth objects and n is the
                        product of image width and height.
        mask2 (torch.Tensor): A tensor of shape (M, n) where M is the number of predicted objects and n is the
                        product of image width and height.
        eps (float, optional): A small value to avoid division by zero.

    Returns:
        (torch.Tensor): A tensor of shape (N, M) representing masks IoU.
    """
    intersection = torch.matmul(mask1, mask2.T).clamp_(0)
    union = (mask1.sum(1)[:, None] + mask2.sum(1)[None]) - intersection  # (area1 + area2) - intersection
    return intersection / (union + eps)


def kpt_iou(
    kpt1: torch.Tensor, kpt2: torch.Tensor, area: torch.Tensor, sigma: List[float], eps: float = 1e-7
) -> torch.Tensor:
    """
    Calculate Object Keypoint Similarity (OKS).

    Args:
        kpt1 (torch.Tensor): A tensor of shape (N, 17, 3) representing ground truth keypoints.
        kpt2 (torch.Tensor): A tensor of shape (M, 17, 3) representing predicted keypoints.
        area (torch.Tensor): A tensor of shape (N,) representing areas from ground truth.
        sigma (list): A list containing 17 values representing keypoint scales.
        eps (float, optional): A small value to avoid division by zero.

    Returns:
        (torch.Tensor): A tensor of shape (N, M) representing keypoint similarities.
    """
    d = (kpt1[:, None, :, 0] - kpt2[..., 0]).pow(2) + (kpt1[:, None, :, 1] - kpt2[..., 1]).pow(2)  # (N, M, 17)
    sigma = torch.tensor(sigma, device=kpt1.device, dtype=kpt1.dtype)  # (17, )
    kpt_mask = kpt1[..., 2] != 0  # (N, 17)
    e = d / ((2 * sigma).pow(2) * (area[:, None, None] + eps) * 2)  # from cocoeval
    # e = d / ((area[None, :, None] + eps) * sigma) ** 2 / 2  # from formula
    return ((-e).exp() * kpt_mask[:, None]).sum(-1) / (kpt_mask.sum(-1)[:, None] + eps)


def _get_covariance_matrix(boxes: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Generate covariance matrix from oriented bounding boxes.

    Args:
        boxes (torch.Tensor): A tensor of shape (N, 5) representing rotated bounding boxes, with xywhr format.

    Returns:
        (torch.Tensor): Covariance matrices corresponding to original rotated bounding boxes.
    """
    # Gaussian bounding boxes, ignore the center points (the first two columns) because they are not needed here.
    gbbs = torch.cat((boxes[:, 2:4].pow(2) / 12, boxes[:, 4:]), dim=-1)
    a, b, c = gbbs.split(1, dim=-1)
    cos = c.cos()
    sin = c.sin()
    cos2 = cos.pow(2)
    sin2 = sin.pow(2)
    return a * cos2 + b * sin2, a * sin2 + b * cos2, (a - b) * cos * sin


def probiou(obb1: torch.Tensor, obb2: torch.Tensor, CIoU: bool = False, eps: float = 1e-7) -> torch.Tensor:
    """
    Calculate probabilistic IoU between oriented bounding boxes.

    Args:
        obb1 (torch.Tensor): Ground truth OBBs, shape (N, 5), format xywhr.
        obb2 (torch.Tensor): Predicted OBBs, shape (N, 5), format xywhr.
        CIoU (bool, optional): If True, calculate CIoU.
        eps (float, optional): Small value to avoid division by zero.

    Returns:
        (torch.Tensor): OBB similarities, shape (N,).

    Notes:
        OBB format: [center_x, center_y, width, height, rotation_angle].

    References:
        https://arxiv.org/pdf/2106.06072v1.pdf
    """
    x1, y1 = obb1[..., :2].split(1, dim=-1)
    x2, y2 = obb2[..., :2].split(1, dim=-1)
    a1, b1, c1 = _get_covariance_matrix(obb1)
    a2, b2, c2 = _get_covariance_matrix(obb2)

    t1 = (
        ((a1 + a2) * (y1 - y2).pow(2) + (b1 + b2) * (x1 - x2).pow(2)) / ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2) + eps)
    ) * 0.25
    t2 = (((c1 + c2) * (x2 - x1) * (y1 - y2)) / ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2) + eps)) * 0.5
    t3 = (
        ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2))
        / (4 * ((a1 * b1 - c1.pow(2)).clamp_(0) * (a2 * b2 - c2.pow(2)).clamp_(0)).sqrt() + eps)
        + eps
    ).log() * 0.5
    bd = (t1 + t2 + t3).clamp(eps, 100.0)
    hd = (1.0 - (-bd).exp() + eps).sqrt()
    iou = 1 - hd
    if CIoU:  # only include the wh aspect ratio part
        w1, h1 = obb1[..., 2:4].split(1, dim=-1)
        w2, h2 = obb2[..., 2:4].split(1, dim=-1)
        v = (4 / math.pi**2) * ((w2 / h2).atan() - (w1 / h1).atan()).pow(2)
        with torch.no_grad():
            alpha = v / (v - iou + (1 + eps))
        return iou - v * alpha  # CIoU
    return iou


def batch_probiou(
    obb1: Union[torch.Tensor, np.ndarray], obb2: Union[torch.Tensor, np.ndarray], eps: float = 1e-7
) -> torch.Tensor:
    """
    Calculate the probabilistic IoU between oriented bounding boxes.

    Args:
        obb1 (torch.Tensor | np.ndarray): A tensor of shape (N, 5) representing ground truth obbs, with xywhr format.
        obb2 (torch.Tensor | np.ndarray): A tensor of shape (M, 5) representing predicted obbs, with xywhr format.
        eps (float, optional): A small value to avoid division by zero.

    Returns:
        (torch.Tensor): A tensor of shape (N, M) representing obb similarities.

    References:
        https://arxiv.org/pdf/2106.06072v1.pdf
    """
    obb1 = torch.from_numpy(obb1) if isinstance(obb1, np.ndarray) else obb1
    obb2 = torch.from_numpy(obb2) if isinstance(obb2, np.ndarray) else obb2

    x1, y1 = obb1[..., :2].split(1, dim=-1)
    x2, y2 = (x.squeeze(-1)[None] for x in obb2[..., :2].split(1, dim=-1))
    a1, b1, c1 = _get_covariance_matrix(obb1)
    a2, b2, c2 = (x.squeeze(-1)[None] for x in _get_covariance_matrix(obb2))

    t1 = (
        ((a1 + a2) * (y1 - y2).pow(2) + (b1 + b2) * (x1 - x2).pow(2)) / ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2) + eps)
    ) * 0.25
    t2 = (((c1 + c2) * (x2 - x1) * (y1 - y2)) / ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2) + eps)) * 0.5
    t3 = (
        ((a1 + a2) * (b1 + b2) - (c1 + c2).pow(2))
        / (4 * ((a1 * b1 - c1.pow(2)).clamp_(0) * (a2 * b2 - c2.pow(2)).clamp_(0)).sqrt() + eps)
        + eps
    ).log() * 0.5
    bd = (t1 + t2 + t3).clamp(eps, 100.0)
    hd = (1.0 - (-bd).exp() + eps).sqrt()
    return 1 - hd


def smooth_bce(eps: float = 0.1) -> Tuple[float, float]:
    """
    Compute smoothed positive and negative Binary Cross-Entropy targets.

    Args:
        eps (float, optional): The epsilon value for label smoothing.

    Returns:
        pos (float): Positive label smoothing BCE target.
        neg (float): Negative label smoothing BCE target.

    References:
        https://github.com/ultralytics/yolov3/issues/238#issuecomment-598028441
    """
    return 1.0 - 0.5 * eps, 0.5 * eps





def smooth(y: np.ndarray, f: float = 0.05) -> np.ndarray:
    """Box filter of fraction f."""
    nf = round(len(y) * f * 2) // 2 + 1  # number of filter elements (must be odd)
    p = np.ones(nf // 2)  # ones padding
    yp = np.concatenate((p * y[0], y, p * y[-1]), 0)  # y padded
    return np.convolve(yp, np.ones(nf) / nf, mode="valid")  # y-smoothed






def ap_per_class(
    tp: np.ndarray,
    conf: np.ndarray,
    pred_cls: np.ndarray,
    target_cls: np.ndarray,
    plot: bool = False,
    on_plot=None,
    save_dir: Path = Path(),
    names: Dict[int, str] = {},
    eps: float = 1e-16,
    prefix: str = "",
) -> Tuple:
    """
    Compute the average precision per class for object detection evaluation.

    Args:
        tp (np.ndarray): Binary array indicating whether the detection is correct (True) or not (False).
        conf (np.ndarray): Array of confidence scores of the detections.
        pred_cls (np.ndarray): Array of predicted classes of the detections.
        target_cls (np.ndarray): Array of true classes of the detections.
        plot (bool, optional): Whether to plot PR curves or not.
        on_plot (callable, optional): A callback to pass plots path and data when they are rendered.
        save_dir (Path, optional): Directory to save the PR curves.
        names (Dict[int, str], optional): Dictionary of class names to plot PR curves.
        eps (float, optional): A small value to avoid division by zero.
        prefix (str, optional): A prefix string for saving the plot files.

    Returns:
        tp (np.ndarray): True positive counts at threshold given by max F1 metric for each class.
        fp (np.ndarray): False positive counts at threshold given by max F1 metric for each class.
        p (np.ndarray): Precision values at threshold given by max F1 metric for each class.
        r (np.ndarray): Recall values at threshold given by max F1 metric for each class.
        f1 (np.ndarray): F1-score values at threshold given by max F1 metric for each class.
        ap (np.ndarray): Average precision for each class at different IoU thresholds.
        unique_classes (np.ndarray): An array of unique classes that have data.
        p_curve (np.ndarray): Precision curves for each class.
        r_curve (np.ndarray): Recall curves for each class.
        f1_curve (np.ndarray): F1-score curves for each class.
        x (np.ndarray): X-axis values for the curves.
        prec_values (np.ndarray): Precision values at mAP@0.5 for each class.
    """
    # Sort by objectness
    i = np.argsort(-conf)
    tp, conf, pred_cls = tp[i], conf[i], pred_cls[i]

    # Find unique classes
    unique_classes, nt = np.unique(target_cls, return_counts=True)
    nc = unique_classes.shape[0]  # number of classes, number of detections

    # Create Precision-Recall curve and compute AP for each class
    x, prec_values = np.linspace(0, 1, 1000), []

    # Average precision, precision and recall curves
    ap, p_curve, r_curve = np.zeros((nc, tp.shape[1])), np.zeros((nc, 1000)), np.zeros((nc, 1000))
    for ci, c in enumerate(unique_classes):
        i = pred_cls == c
        n_l = nt[ci]  # number of labels
        n_p = i.sum()  # number of predictions
        if n_p == 0 or n_l == 0:
            continue

        # Accumulate FPs and TPs
        fpc = (1 - tp[i]).cumsum(0)
        tpc = tp[i].cumsum(0)

        # Recall
        recall = tpc / (n_l + eps)  # recall curve
        r_curve[ci] = np.interp(-x, -conf[i], recall[:, 0], left=0)  # negative x, xp because xp decreases

        # Precision
        precision = tpc / (tpc + fpc)  # precision curve
        p_curve[ci] = np.interp(-x, -conf[i], precision[:, 0], left=1)  # p at pr_score

        # AP from recall-precision curve
        for j in range(tp.shape[1]):
            ap[ci, j], mpre, mrec = compute_ap(recall[:, j], precision[:, j])
            if j == 0:
                prec_values.append(np.interp(x, mrec, mpre))  # precision at mAP@0.5

    prec_values = np.array(prec_values) if prec_values else np.zeros((1, 1000))  # (nc, 1000)

    # Compute F1 (harmonic mean of precision and recall)
    f1_curve = 2 * p_curve * r_curve / (p_curve + r_curve + eps)
    names = {i: names[k] for i, k in enumerate(unique_classes) if k in names}  # dict: only classes that have data
    if plot:
        plot_pr_curve(x, prec_values, ap, save_dir / f"{prefix}PR_curve.png", names, on_plot=on_plot)
        plot_mc_curve(x, f1_curve, save_dir / f"{prefix}F1_curve.png", names, ylabel="F1", on_plot=on_plot)
        plot_mc_curve(x, p_curve, save_dir / f"{prefix}P_curve.png", names, ylabel="Precision", on_plot=on_plot)
        plot_mc_curve(x, r_curve, save_dir / f"{prefix}R_curve.png", names, ylabel="Recall", on_plot=on_plot)

    i = smooth(f1_curve.mean(0), 0.1).argmax()  # max F1 index
    p, r, f1 = p_curve[:, i], r_curve[:, i], f1_curve[:, i]  # max-F1 precision, recall, F1 values
    tp = (r * nt).round()  # true positives
    fp = (tp / (p + eps) - tp).round()  # false positives
    return tp, fp, p, r, f1, ap, unique_classes.astype(int), p_curve, r_curve, f1_curve, x, prec_values






import pytest
import numpy as np
import imageio.v3 as iio
from pathlib import Path
from pycvt import draw_bounding_boxes, overlay_masks
from cvmd import Yolov5Detect, Yolov5Segment, Yolov8Segment, Yolov11Detect, Yolov11Segment, DETR

def run_detection_test(model_class, weights, coco_images, blank_image, save_dir, name, imgsz=640):
    if not Path(weights).exists():
        pytest.skip(f"Weights not found: {weights}")
    
    model = model_class(
        weights=weights,
        device="cuda",
        conf=0.25,
        iou=0.45,
        imgsz=imgsz,
        half=True,
    )
    model.load_model()
    
    model_save_dir = save_dir / name
    model_save_dir.mkdir(parents=True, exist_ok=True)

    # Test real images
    for img_path in coco_images:
        sample_image = iio.imread(img_path, mode="RGB")
        results = model(sample_image)
        bbox = results[..., :4]
        conf = results[..., 4]
        cls = results[..., 5].astype(int)
        plot_im = draw_bounding_boxes(
            image=sample_image,
            boxes=bbox,
            labels=[f"{c}-{s:.2f}" for c, s in zip(cls, conf)],
            line_width=2,
        )
        iio.imwrite(model_save_dir / f"{img_path.name}", plot_im)
    
    # Test blank image
    blank_results = model(blank_image)
    assert len(blank_results) >= 0

def run_segmentation_test(model_class, weights, coco_images, blank_image, save_dir, name):
    if not Path(weights).exists():
        pytest.skip(f"Weights not found: {weights}")
        
    model = model_class(
        weights=weights,
        device="cuda",
        conf=0.25,
        iou=0.45,
        imgsz=640,
        half=True,
    )
    model.load_model()
    
    model_save_dir = save_dir / name
    model_save_dir.mkdir(parents=True, exist_ok=True)

    # Test real images
    for img_path in coco_images:
        sample_image = iio.imread(img_path, mode="RGB")
        results, masks = model(sample_image)
        plot_im = overlay_masks(image=sample_image, masks=masks)
        
        bbox = results[..., :4]
        conf = results[..., 4]
        cls = results[..., 5].astype(int)
        plot_im = draw_bounding_boxes(
            image=plot_im,
            boxes=bbox,
            labels=[f"{c}-{s:.2f}" for c, s in zip(cls, conf)],
            line_width=2,
        )
        iio.imwrite(model_save_dir / f"{img_path.name}", plot_im)
    
    # Test blank image
    blank_results, blank_masks = model(blank_image)
    assert len(blank_results) >= 0

@pytest.mark.parametrize("model_info", [
    (Yolov5Detect, "temp/model_weights/yolov5l.torchscript", "yolov5l_detect", 640),
    (Yolov11Detect, "temp/model_weights/yolo11l.torchscript", "yolov11l_detect", 640),
    (DETR, "temp/model_weights/detr_resnet50.torchscript", "detr_resnet50", 800),
])
def test_detection(model_info, coco_images, blank_image, save_dir):
    model_class, weights, name, imgsz = model_info
    run_detection_test(model_class, weights, coco_images, blank_image, save_dir, name, imgsz=imgsz)

@pytest.mark.parametrize("model_info", [
    (Yolov5Segment, "temp/model_weights/yolov5l-seg.torchscript", "yolov5l_seg"),
    (Yolov8Segment, "temp/model_weights/yolov8l-seg.torchscript", "yolov8l_seg"),
    (Yolov11Segment, "temp/model_weights/yolo11l-seg.torchscript", "yolov11l_seg"),
])
def test_segmentation(model_info, coco_images, blank_image, save_dir):
    model_class, weights, name = model_info
    run_segmentation_test(model_class, weights, coco_images, blank_image, save_dir, name)



from typing import Callable, Dict, Any, Iterable, List, Optional


from cvmd import build


try:
    # only import ray when needed
    import ray
    from ray.util.actor_pool import ActorPool
except ImportError:
    ray = None
    ActorPool = None


class InferActor:
    def __init__(
        self,
        runs_config: Dict[str, Any] = {},
        model_config: Dict[str, Any] = {},
        handler: Optional[Callable[..., Any]] = None,
    ):
        # print("Initializing InferActor...")
        # print(f"Model config: {model_config}")
        # print(f"Runs config: {runs_config}")
        self.model = build(
            model_config.get("model_name"),
            **model_config,
        )
        self.model.load_model()
        self.runs_config = runs_config
        self.model_config = model_config
        self.model_config["model"] = self.model
        self.handler = handler

    def infer(
        self,
        task: Any,
        *args,
        **kwds,
    ) -> Dict[str, Any]:
        return self.handler(task, self.model_config, self.runs_config, *args, **kwds)


def ray_infer_iter(
    InferActorCls,
    tasks: List[Any],
    *,
    num_actors: Optional[int] = None,
    num_cpus: float = 2.0,
    gpus_per_actor: float = 0.25,
    actor_kwargs: Optional[dict] = None,
    remote_method: str = "infer",
    remote_args: Optional[tuple] = None,
    remote_kwargs: Optional[dict] = None,
) -> Iterable[Any]:

    if not ray.is_initialized():
        ray.init()

    actor_kwargs = actor_kwargs or {}
    remote_args = remote_args or ()
    remote_kwargs = remote_kwargs or {}

    total_gpus = ray.cluster_resources().get("GPU", 0)
    if total_gpus > 0:
        if num_actors is None:
            # Standard Ray pattern: Fix resource requirement per actor, 
            # and dynamically calculate the number of actors to fill the cluster.
            num_actors = int(total_gpus / gpus_per_actor)
    else:
        gpus_per_actor = 0
        if num_actors is None:
            num_actors = 1

    num_actors = max(1, num_actors)

    RemoteInfer = ray.remote(num_gpus=gpus_per_actor, num_cpus=num_cpus)(InferActorCls)
    actors = [RemoteInfer.remote(**actor_kwargs) for _ in range(num_actors)]
    pool = ActorPool(actors)

    for r in pool.map_unordered(
        lambda a, t: getattr(a, remote_method).remote(t, *remote_args, **remote_kwargs),
        tasks,
    ):
        yield r



from .yolo.yolov5 import Yolov5Detect as Yolov5Detect, Yolov5Segment as Yolov5Segment
from .yolo.yolov8 import Yolov8Detect as Yolov8Detect, Yolov8Segment as Yolov8Segment, YoloInitKwargs as YoloInitKwargs
from .yolo.yolov11 import Yolov11Detect as Yolov11Detect, Yolov11Segment as Yolov11Segment
from .yolo.yolov12 import Yolov12Detect as Yolov12Detect, Yolov12Segment as Yolov12Segment
from .detr.detr import DETR as DETR
from .detr.deformable_detr import DeformableDETR as DeformableDETR
from .utils.windows import detect_with_windows as detect_with_windows
from .utils.imageutils import IMAGE_EXTS as IMAGE_EXTS
from .registry import register_model as register_model, list_models as list_models, build as build

__all__ = [
    "DETR",
    "DeformableDETR",
    "Yolov5Detect",
    "Yolov5Segment",
    "YoloInitKwargs",
    "Yolov8Detect",
    "Yolov8Segment",
    "Yolov11Detect",
    "Yolov11Segment",
    "Yolov12Detect",
    "Yolov12Segment",
    "detect_with_windows",
    "IMAGE_EXTS",
    "register_model",
    "list_models",
    "build",
]



import time

import torch
import torchvision
from torchvision.ops import box_iou
import torch.nn.functional as F
from .ops import xywh2xyxy


def non_max_suppression(
    prediction,
    conf_thres=0.25,
    iou_thres=0.45,
    classes=None,
    agnostic=False,
    multi_label=False,
    labels=(),
    max_det=300,
    nm=0,  # number of masks
):
    """Non-Maximum Suppression (NMS) on inference results to reject overlapping detections.

    Returns:
        list of detections, on (n,6) tensor per image [xyxy, conf, cls]
    """
    # Checks
    assert 0 <= conf_thres <= 1, f"Invalid Confidence threshold {conf_thres}, valid values are between 0.0 and 1.0"
    assert 0 <= iou_thres <= 1, f"Invalid IoU {iou_thres}, valid values are between 0.0 and 1.0"
    if isinstance(prediction, (list, tuple)):  # YOLOv5 model in validation model, output = (inference_out, loss_out)
        prediction = prediction[0]  # select only inference output

    device = prediction.device
    mps = "mps" in device.type  # Apple MPS
    if mps:  # MPS not fully supported yet, convert tensors to CPU before NMS
        prediction = prediction.cpu()
    bs = prediction.shape[0]  # batch size
    nc = prediction.shape[2] - nm - 5  # number of classes
    xc = prediction[..., 4] > conf_thres  # candidates

    # Settings
    # min_wh = 2  # (pixels) minimum box width and height
    max_wh = 7680  # (pixels) maximum box width and height
    max_nms = 30000  # maximum number of boxes into torchvision.ops.nms()
    time_limit = 0.5 + 0.05 * bs  # seconds to quit after
    redundant = True  # require redundant detections
    multi_label &= nc > 1  # multiple labels per box (adds 0.5ms/img)
    merge = False  # use merge-NMS

    t = time.time()
    mi = 5 + nc  # mask start index
    output = [torch.zeros((0, 6 + nm), device=prediction.device)] * bs
    for xi, x in enumerate(prediction):  # image index, image inference
        # Apply constraints
        # x[((x[..., 2:4] < min_wh) | (x[..., 2:4] > max_wh)).any(1), 4] = 0  # width-height
        x = x[xc[xi]]  # confidence

        # Cat apriori labels if autolabelling
        if labels and len(labels[xi]):
            lb = labels[xi]
            v = torch.zeros((len(lb), nc + nm + 5), device=x.device)
            v[:, :4] = lb[:, 1:5]  # box
            v[:, 4] = 1.0  # conf
            v[range(len(lb)), lb[:, 0].long() + 5] = 1.0  # cls
            x = torch.cat((x, v), 0)

        # If none remain process next image
        if not x.shape[0]:
            continue

        # Compute conf
        x[:, 5:] *= x[:, 4:5]  # conf = obj_conf * cls_conf

        # Box/Mask
        box = xywh2xyxy(x[:, :4])  # center_x, center_y, width, height) to (x1, y1, x2, y2)
        mask = x[:, mi:]  # zero columns if no masks

        # Detections matrix nx6 (xyxy, conf, cls)
        if multi_label:
            i, j = (x[:, 5:mi] > conf_thres).nonzero(as_tuple=False).T
            x = torch.cat((box[i], x[i, 5 + j, None], j[:, None].float(), mask[i]), 1)
        else:  # best class only
            conf, j = x[:, 5:mi].max(1, keepdim=True)
            x = torch.cat((box, conf, j.float(), mask), 1)[conf.view(-1) > conf_thres]

        # Filter by class
        if classes is not None:
            x = x[(x[:, 5:6] == torch.tensor(classes, device=x.device)).any(1)]

        # Apply finite constraint
        # if not torch.isfinite(x).all():
        #     x = x[torch.isfinite(x).all(1)]

        # Check shape
        n = x.shape[0]  # number of boxes
        if not n:  # no boxes
            continue
        x = x[x[:, 4].argsort(descending=True)[:max_nms]]  # sort by confidence and remove excess boxes

        # Batched NMS
        c = x[:, 5:6] * (0 if agnostic else max_wh)  # classes
        boxes, scores = x[:, :4] + c, x[:, 4]  # boxes (offset by class), scores
        i = torchvision.ops.nms(boxes, scores, iou_thres)  # NMS
        i = i[:max_det]  # limit detections
        if merge and (1 < n < 3e3):  # Merge NMS (boxes merged using weighted mean)
            # update boxes as boxes(i,4) = weights(i,n) * boxes(n,4)
            iou = box_iou(boxes[i], boxes) > iou_thres  # iou matrix
            weights = iou * scores[None]  # box weights
            x[i, :4] = torch.mm(weights, x[:, :4]).float() / weights.sum(1, keepdim=True)  # merged boxes
            if redundant:
                i = i[iou.sum(1) > 1]  # require redundancy

        output[xi] = x[i]
        if mps:
            output[xi] = output[xi].to(device)
        if (time.time() - t) > time_limit:
            print(f"WARNING ⚠️ NMS time limit {time_limit:.3f}s exceeded")
            break  # time limit exceeded

    return output


def crop_mask(masks, boxes):
    """Crop predicted masks by zeroing out everything not in the predicted bbox.

    Args:
        - masks should be a size [n, h, w] tensor of masks
        - boxes should be a size [n, 4] tensor of bbox coords in relative point form.
    """
    _n, h, w = masks.shape
    x1, y1, x2, y2 = torch.chunk(boxes[:, :, None], 4, 1)  # x1 shape(1,1,n)
    r = torch.arange(w, device=masks.device, dtype=x1.dtype)[None, None, :]  # rows shape(1,w,1)
    c = torch.arange(h, device=masks.device, dtype=x1.dtype)[None, :, None]  # cols shape(h,1,1)

    return masks * ((r >= x1) * (r < x2) * (c >= y1) * (c < y2))


def process_mask_native(protos, masks_in, bboxes, shape):
    """Crop after upsample.

    Args:
        protos: [mask_dim, mask_h, mask_w]
        masks_in: [n, mask_dim], n is number of masks after nms
        bboxes: [n, 4], n is number of masks after nms
        shape: input_image_size, (h, w).

    Returns:
        h, w, n
    """
    c, mh, mw = protos.shape  # CHW
    masks = (masks_in @ protos.float().view(c, -1)).sigmoid().view(-1, mh, mw)
    gain = min(mh / shape[0], mw / shape[1])  # gain  = old / new
    pad = (mw - shape[1] * gain) / 2, (mh - shape[0] * gain) / 2  # wh padding
    top, left = int(pad[1]), int(pad[0])  # y, x
    bottom, right = int(mh - pad[1]), int(mw - pad[0])
    masks = masks[:, top:bottom, left:right]

    masks = F.interpolate(masks[None], shape, mode="bilinear", align_corners=False)[0]  # CHW
    masks = crop_mask(masks, bboxes)  # CHW
    return masks.gt_(0.5)


from __future__ import annotations
from typing import Dict, Type, Callable, Any

_MODEL_REGISTRY: Dict[str, Type] = {}   # lower(name) -> class


def register_model(*names: str) -> Callable[[Type], Type]:
    if not names:
        raise ValueError("register_model() requires at least one name")

    def _wrap(cls: Type) -> Type:
        for n in names:
            key = n.lower()
            if key in _MODEL_REGISTRY and _MODEL_REGISTRY[key] is not cls:
                raise KeyError(
                    f"Model name '{n}' already registered by {_MODEL_REGISTRY[key]}"
                )
            _MODEL_REGISTRY[key] = cls

        cls.MODEL_NAMES = tuple(sorted({n.lower() for n in names}))

        # ✅ 正确的打印位置：这里才能看到 cls
        # print(f"Registering model: {names} -> {cls}")

        return cls

    return _wrap


def build(model: str | Type, *args: Any, **kwargs: Any):
    """
    build("yolov8det", weights=..., device=...)
    build(Yolov8Detect, weights=..., device=...)
    """
    if isinstance(model, str):
        key = model.lower()
        if key not in _MODEL_REGISTRY:
            raise KeyError(f"Unknown model '{model}'. Available: {list_models()}")
        cls = _MODEL_REGISTRY[key]
    else:
        cls = model
    # print(f"Building model: {cls}")
    return cls(*args, **kwargs)


def list_models():
    """Return all registered names (lowercased)."""
    return sorted(_MODEL_REGISTRY.keys())



import numpy as np


def detect_with_windows(
    image, windows, model, merge=True, merge_iou=0.2
):
    """
    Detect in multiple windows and optionally merge overlapped duplicates.

    model(crop) -> results (N,6) or (results, extra)
    results format: [x1,y1,x2,y2, conf, cls] in window-local coords

    Args:
        image: (H,W,3) np.ndarray
        windows: list of (x1,y1,x2,y2)
        model: function that takes in an image crop and returns detections
        merge: using WBF to merge boxes from different windows
        merge_iou: IoU threshold for merging boxes,default 0.001 to merge almost all overlapping boxes
        class_aware: whether to merge boxes only of the same class
    """
    H, W = image.shape[:2]

    windows = np.asarray(windows).astype(int)
    windows = windows.clip([0, 0, 0, 0], [W, H, W, H])

    part_imgs = [image[y1:y2, x1:x2] for (x1, y1, x2, y2) in windows]
    all_results = []
    for crop, (wx1, wy1, wx2, wy2) in zip(part_imgs, windows):
        results = model(crop)
        pred = results if isinstance(results, np.ndarray) else results[0]
        if pred.shape[0] > 0:
            pred[:, :4] += np.tile((wx1, wy1), 2)
            all_results.append(pred)

    pred = np.vstack(all_results) if all_results else np.empty((0, 6))

    if not merge:
        return pred
    else:
        from ensemble_boxes import weighted_boxes_fusion
        scale = np.array([W, H, W, H], dtype=np.float32)
        pred[:, :4] /= scale
        
        boxes, scores, labels = weighted_boxes_fusion(
            boxes_list=[pred[:, :4].tolist()],
            scores_list=[pred[:, 4].tolist()],
            labels_list=[pred[:, 5].astype(int).tolist()],
            iou_thr=merge_iou
        )
        boxes = np.asarray(boxes, dtype=np.float32) * scale
        scores = np.asarray(scores)
        labels = np.asarray(labels)
        return np.hstack((boxes, scores[:, None], labels[:, None]))



from cvmd.registry import register_model
from .yolov8 import Yolov8Detect, Yolov8Segment

@register_model("yolov12", "yolov12det", "yolov12detect")
class Yolov12Detect(Yolov8Detect):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

@register_model("yolov12seg", "yolov12segment")
class Yolov12Segment(Yolov8Segment):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from typing import Tuple, TypedDict, Optional, Sequence, Union

import torch

from cvmd.registry import register_model
from cvmd.utils.torchutils import normalize_device

from .ops import (
    letterbox,
    non_max_suppression,
    process_mask_native,
    scale_boxes,
)


class YoloInitKwargs(TypedDict, total=False):
    model_name: Optional[str]
    weights: Union[str, bytes, bytearray]
    device: Union[str, torch.device]
    conf: float
    iou: float
    classes: Optional[Sequence[int]]
    imgsz: Union[int, Tuple[int, int]]
    half: bool
    nc: Optional[int]
    load_warm_up: bool
    mask_thr: float

@register_model("yolov8", "yolov8det", "yolov8detect")
class Yolov8Detect:

    def __init__(self, *args, **kwargs: YoloInitKwargs):
        self.weights = kwargs.get("weights", "yolov8s.torchscript")
        self.device = normalize_device(kwargs.get("device", "cpu"))
        self.conf = kwargs.get("conf", 0.25)
        self.iou = kwargs.get("iou", 0.45)
        self.classes = kwargs.get("classes", None)
        self.imgsz = kwargs.get("imgsz", 640)
        self.half = kwargs.get("half", False)
        self.nc = kwargs.get("nc", None)
        self._load_warmup = kwargs.get("load_warm_up", True)
        self._model_dtype = None

        if isinstance(self.imgsz, int):
            self.imgsz = (self.imgsz, self.imgsz)

    def load_model(self, *args, **kwds):
        self.weights = kwds.get("weights", self.weights)
        self.device = normalize_device(kwds.get("device", self.device))
        if isinstance(self.weights, (bytes, bytearray)):
            import io

            self.weights = io.BytesIO(self.weights)
        model = torch.jit.load(self.weights, map_location=self.device)
        model.eval()
        use_half = self.half and self.device.type != "cpu"
        self.model = model.half() if use_half else model.float()
        self._model_dtype = next(self.model.parameters()).dtype
        self._warmup() if self._load_warmup else None
    

    def _warmup(self, *args, **kwds):
        import numpy as np

        empty_im = np.zeros((*self.imgsz, 3), dtype=np.uint8)  # HWC
        self.__call__(empty_im)

    def __call__(self, *args, **kwds):
        """
        arguments:
            images: numpy array of shape (h,w,3) in RGB format

        returns: numpy array of shape (n,6) where each row is
            x1, y1, x2, y2, confidence, class
        """
        if not hasattr(self, "model"):
            raise RuntimeError("Model is not loaded. Call `load_model` first.")

        with torch.inference_mode():
            x = self.__pre_process__(*args, **kwds)
            pred = self.model(x)
            pred = self.__post_process__(pred, *args, **kwds)
        return pred

    def __post_process__(self, pred, *args, **kwds):
        pred = non_max_suppression(
            pred,
            self.conf,
            self.iou,
            classes=self.classes,
            nc=self.nc,
        )[0]
        pred[:, :4] = scale_boxes(self.imgsz, pred[:, :4], args[0].shape[:2]).round()
        return pred.cpu().numpy()

    def __pre_process__(self, *args, **kwds):
        im0 = args[0]
        im1 = letterbox(im0, new_shape=self.imgsz)
        x = torch.from_numpy(im1).permute(2, 0, 1).to(self.device)
        x = x.to(self._model_dtype)
        x = x.unsqueeze(0) / 255.0
        return x

@register_model("yolov8seg", "yolov8segment")
class Yolov8Segment(Yolov8Detect):

    def __init__(self, *args, **kwargs: YoloInitKwargs):
        self.mask_thr = kwargs.pop("mask_thr", 0.6)
        self.nc = kwargs.get("nc", None)
        super().__init__(*args, **kwargs)

    def __post_process__(self, pred, *args, **kwds):
        im0 = args[0]
        pred0, proto = pred

        if self.nc is None:
            self.nc = int(pred0.shape[1] - 4 - proto.shape[1])

        proto = proto[0]
        det = non_max_suppression(
            pred0, self.conf, self.iou, classes=self.classes, nc=self.nc
        )[0]
        if det is None or len(det) == 0:
            import numpy as np
            return np.zeros((0, 6), dtype=np.float32), np.zeros((0, *im0.shape[:2]), dtype=bool)

        det[:, :4] = scale_boxes(self.imgsz, det[:, :4], im0.shape[:2]).round()
        masks = process_mask_native(proto, det[:, 6:], det[:, :4], im0.shape[:2])
        masks = masks > self.mask_thr

        return det[:, :6].cpu().numpy(), masks.cpu().numpy()
